from ast import literal_eval
from functools import lru_cache
from os import getenv, path
from subprocess import CalledProcessError, STDOUT, check_output
from textwrap import dedent
import logging
import re
import sys

from .fdio import tmpdir

__all__ = ["get_engine"]
IS_WIN = any(sys.platform.startswith(i) for i in ["win32", "cygwin"])
MATLAB_RUN = "matlab -nodesktop -nosplash -nojvm".split()
if IS_WIN:
    MATLAB_RUN += ["-wait", "-log"]
log = logging.getLogger(__name__)


class VersionError(ValueError):
    pass


def check_output_u8(*args, **kwargs):
    return check_output(*args, **kwargs).decode("utf-8").strip()


@lru_cache()
def get_engine(name=None):
    try:
        from matlab import engine
    except ImportError:
        try:
            log.warn(
                dedent(
                    """\
                Python could not find the MATLAB engine.
                Attempting to install automatically."""
                )
            )
            log.debug(_install_engine())
            log.info("installed MATLAB engine for Python")
            from matlab import engine
        except CalledProcessError:
            raise ImportError(
                dedent(
                    """\
                Please install MATLAB and its Python module.
                See https://www.mathworks.com/help/matlab/matlab_external/\
install-the-matlab-engine-for-python.html
                or
                https://www.mathworks.com/help/matlab/matlab_external/\
install-matlab-engine-api-for-python-in-nondefault-locations.html
                It's likely you need to do:

                cd "{matlabroot}\\extern\\engines\\python"
                {exe} setup.py build --build-base="BUILDDIR" install

                - Fill in any temporary directory name for BUILDDIR
                  (e.g. /tmp/builddir).
                - If installation fails due to write permissions, try appending `--user`
                  to the above command.
                """
                ).format(
                    matlabroot=matlabroot(default="matlabroot"), exe=sys.executable
                )
            )
    started = engine.find_matlab()
    notify = False
    if not started or (name and name not in started):
        notify = True
        log.debug("Starting MATLAB")
    eng = engine.connect_matlab(name=name or getenv("SPM12_MATLAB_ENGINE", None))
    if notify:
        log.debug("MATLAB started")
    return eng


def _matlab_run(command, jvm=False, auto_exit=True):
    if auto_exit and not command.endswith("exit"):
        command = command + ", exit"
    return check_output_u8(
        MATLAB_RUN + ([] if jvm else ["-nojvm"]) + ["-r", command], stderr=STDOUT
    )


def matlabroot(default=None):
    if IS_WIN:
        try:
            res = _matlab_run("display(matlabroot);")
        except CalledProcessError:
            if default:
                return default
            raise
        return re.search(r"^([A-Z]:\\.*)\s*$", res, flags=re.M).group(1)

    try:
        res = check_output_u8(["matlab", "-n"])
    except CalledProcessError:
        if default:
            return default
        raise
    return re.search(r"MATLAB\s+=\s+(\S+)\s*$", res, flags=re.M).group(1)


def _install_engine():
    src = path.join(matlabroot(), "extern", "engines", "python")
    with open(path.join(src, "setup.py")) as fd:  # check version support
        supported = literal_eval(
            re.search(r"supported_version.*?=\s*(.*?)$", fd.read(), flags=re.M).group(1)
        )
        if ".".join(map(str, sys.version_info[:2])) not in map(str, supported):
            raise VersionError(
                dedent(
                    """\
                Python version is {info[0]}.{info[1]},
                but the installed MATLAB only supports Python versions: [{supported}]
                """.format(
                        info=sys.version_info[:2], supported=", ".join(supported)
                    )
                )
            )
    with tmpdir() as td:
        cmd = [sys.executable, "setup.py", "build", "--build-base", td, "install"]
        try:
            return check_output_u8(cmd, cwd=src)
        except CalledProcessError:
            log.warn("Normal install failed. Attempting `--user` install.")
            return check_output_u8(cmd + ["--user"], cwd=src)



from pytest import importorskip

from miutil.imio import imread


def test_imread(tmp_path):
    np = importorskip("numpy")

    x = np.random.randint(10, size=(9, 9))
    fname = tmp_path / "test_imread.npy"
    np.save(fname, x)
    assert (imread(str(fname)) == x).all()

    fname = tmp_path / "test_imread.npz"
    np.savez(fname, x)
    assert (imread(str(fname)) == x).all()

    np.savez(fname, x, x)
    assert (imread(str(fname))["arr_0"] == x).all()



from os import path
import matplotlib.pyplot as plt

from .imio import imread


class imscroll:
    """
    Slice through volumes by scrolling.
    Hold SHIFT to scroll faster.
    """

    instances = []
    SUPPORTED_KEYS = ["shift"]

    def __init__(self, vol, view="t", fig=None, titles=None, **kwargs):
        """
        Scroll through 2D slices of 3D volume(s) using the mouse.
        Args:
            vol (str or numpy.ndarray or list or dict): path to file or
                a (list/dict of) array(s).
            view (str): z, t, transverse/y, c, coronal/x, s, sagittal.
            fig (matplotlib.pyplot.Figure): will be created if unspecified.
            titles (list): list of strings (overrides `vol.keys()`).
            **kwargs: passed to `matplotlib.pyplot.imshow()`.
        """
        if isinstance(vol, str) and path.exists(vol):
            vol = imread(vol)
        if hasattr(vol, "keys"):
            keys = list(vol.keys())
            vol = [vol[i] for i in keys]
            if titles is None:
                titles = keys
        if vol[0].ndim == 2:
            vol = [vol]
        elif vol[0].ndim != 3:
            raise IndexError("Expected vol.ndim in [3, 4] but got {}".format(vol.ndim))

        self.titles = titles or [None] * len(vol)

        view = view.lower()
        if view in ["c", "coronal", "y"]:
            vol = [i.transpose(1, 0, 2) for i in vol]
        elif view in ["s", "saggital", "x"]:
            vol = [i.transpose(2, 0, 1) for i in vol]

        self.index_max = min(map(len, vol))
        self.index = self.index_max // 2
        if fig is not None:
            self.fig, axs = fig, fig.subplots(1, len(vol))
        else:
            self.fig, axs = plt.subplots(1, len(vol))
        self.axs = [axs] if len(vol) == 1 else list(axs.flat)
        for ax, i, t in zip(self.axs, vol, self.titles):
            ax.imshow(i[self.index], **kwargs)
            ax.set_title(t or "slice #{}".format(self.index))
        self.vols = vol
        self.key = {i: False for i in self.SUPPORTED_KEYS}
        self.fig.canvas.mpl_connect("scroll_event", self.scroll)
        self.fig.canvas.mpl_connect("key_press_event", self.on_key)
        self.fig.canvas.mpl_connect("key_release_event", self.off_key)
        imscroll.instances.append(self)  # prevents gc

    @classmethod
    def clear(cls, self):
        cls.instances.clear()

    def on_key(self, event):
        if event.key in self.SUPPORTED_KEYS:
            self.key[event.key] = True

    def off_key(self, event):
        if event.key in self.SUPPORTED_KEYS:
            self.key[event.key] = False

    def scroll(self, event):
        self.set_index(
            self.index
            + (1 if event.button == "up" else -1) * (10 if self.key["shift"] else 1)
        )

    def set_index(self, index):
        self.index = index % self.index_max
        for ax, vol, t in zip(self.axs, self.vols, self.titles):
            ax.images[0].set_array(vol[self.index])
            ax.set_title(t or "slice #{}".format(self.index))
        self.fig.canvas.draw()



import re

RE_NII_GZ = re.compile(r"^(.+)(\.nii(?:\.gz)?)$", flags=re.I)
RE_NPYZ = re.compile(r"^(.+)(\.np[yz])$", flags=re.I)


def imread(fname):
    """Read any supported filename"""
    if RE_NII_GZ.search(fname):
        from .nii import getnii

        return getnii(fname)
    elif RE_NPYZ.search(fname):
        import numpy as np

        res = np.load(fname)
        if hasattr(res, "keys") and len(res.keys()) == 1:
            res = res[list(res.keys())[0]]
        return res
    raise ValueError("Unknown image type: " + fname)



from os import chdir, path
import sys
from subprocess import check_call

chdir(path.dirname(path.dirname(path.abspath(__file__))))
check_call(
    [
        sys.executable,
        "-m",
        "pytest",
        "-v",
        "-n=4",
        "-r=xs",
        "--timeout=600",
        "--timeout_method=thread",
        "--log-level=debug",
        "--cov=miutil",
        "--durations=0",
    ]
    + sys.argv[1:],
)
